/*
** Copyright 2000-2007 Double Precision, Inc.
** See COPYING for distribution information.
*/

#include	"config.h"
#include	"rfc822/rfc2047.h"
#include	"unicode/unicode.h"
#include	"cmlm.h"
#include	"cmlmcmdmisc.h"
#include	"cmlmsubunsub.h"
#include	"dbobj.h"
#include	"cmlmsublist.h"
#include	<iostream>
#include	<fstream>
#include	<sstream>
#include	<stdio.h>
#include	<string.h>
#include	<stdlib.h>
#include	<ctype.h>
#include	<time.h>
#include	"mydirent.h"
#include	<fcntl.h>
#include	<sys/types.h>
#include	<sys/stat.h>
#include	<sysexits.h>
#if HAVE_LANGINFO_H
#include	<langinfo.h>
#endif
#if	HAVE_LOCALE_H
#include	<locale.h>
#endif

static const char rcsid[]="$Id: cmlmcmdmisc.C,v 1.10 2007/03/12 00:13:54 mrsam Exp $";

static int updatelistheaders()
{
	std::string addr=cmdget_s("ADDRESS");

	std::string url=cmdget_s("URL");

	if (addr.size() > 0)
	{
		std::ofstream ofs(HEADERADD ".new");

		// Preserve existing non-List headers.

		std::ifstream ifs(HEADERADD);

		if (ifs.is_open())
		{
			std::string line;

			while (!std::getline(ifs, line).eof())
			{
				std::string::iterator
					b=line.begin(), e=line.end();

				std::string p(b, std::find(b, e, ':'));

				std::transform(p.begin(), p.end(),
					       p.begin(),
					       std::ptr_fun(::tolower));

				if (p.substr(0, 5) == "list-")
					continue;

				ofs << line << std::endl;
			}
		}

		ofs << "List-Subscribe: <mailto:" <<
			get_verp_return("subscribe") << ">";

		if (url.size() > 0)
			ofs << ", <" << url << ">";

		ofs << std::endl;

		ofs << "List-Unsubscribe: <mailto:" <<
			get_verp_return("unsubscribe") << ">";

		if (url.size() > 0)
			ofs << ", <" << url << ">";

		ofs << std::endl;

		ofs << "List-Post: <mailto:" <<
			cmdget_s("ADDRESS") << ">" << std::endl
			<< "List-Owner: <mailto:" <<
			get_verp_return("owner") << ">" << std::endl
			<< "List-Help: <mailto:" <<
			get_verp_return("help") << ">" << std::endl;

		ofs.flush();

		if (ofs.fail())
		{
			perror(HEADERADD);
			return 1;
		}
		ofs.close();
		if (rename(HEADERADD ".new", HEADERADD))
			return 1;
	}
	return 0;
}

int cmdupdate(const std::vector<std::string> &argv)
{
	DIR	*dirp;
	struct	dirent *de;
	std::string addr=cmdget_s("ADDRESS");

	// Copy all the template files

	dirp=opendir(TEMPLATEDIR);
	while (dirp && (de=readdir(dirp)) != 0)
	{
		const char *p=strrchr(de->d_name, '.');
		std::string n;
		std::string line;
		std::string s;

		if (!p)
			continue;

		if (strcmp(p, ".tmpl") && strcmp(p, ".html"))
			continue;

		n= TEMPLATEDIR "/";
		n += de->d_name;

		std::ifstream	ifs(n.c_str());

		if (ifs.bad())
		{
			perror(n.c_str());
			return (1);
		}

		std::ofstream	ofs(de->d_name);

		if (ofs.bad())
		{
			perror(de->d_name);
			return (1);
		}

		// Copy line at a time, expanding the address macro

		while (!std::getline(ifs, line).eof())
		{
			size_t i;

			for (i=0; i+1<line.size(); i++)
			{
				if (line[i] != '#' || line[i+1] != '#')
					continue;

				s=line.substr(i+2);

				size_t j=s.find('#');

				if (j == s.npos || j+1 >= s.size() ||
				    s[j+1] != '#')
					continue;

				s=s.substr(0, j);

				bool add_mailto=true;

				if (s.substr(0, 1) == "-")
				{
					s=s.substr(1);
					add_mailto=false;
				}

				if (s.size() > 0)	s= "-" + s;

				size_t k=addr.rfind('@');

				if (k == addr.npos)	k=0;

				s= (add_mailto ? "mailto:":"")
					+ addr.substr(0, k)
					+ s + addr.substr(k);
			
				line=line.substr(0, i) +s+ line.substr(i+j+4);
				i += j+3;	
			}
			ofs << line << std::endl;
		}
		ofs.flush();
		if (ofs.fail())
		{
			perror(de->d_name);
			return (1);
		}
		ofs.close();
		if (ofs.fail())
		{
			perror(de->d_name);
			return (1);
		}
	}
	if (dirp)	closedir(dirp);

	return updatelistheaders();
}

// Create a mailing list

int cmdcreate(const std::vector<std::string> &argv)
{
	std::ofstream ofs(OPTIONS);

	if (!ofs.is_open())
	{
		perror(OPTIONS);
		return (1);
	}

	std::vector<std::string>::const_iterator
		b=argv.begin(),
		e=argv.end();

	while (b != e)
	{
		ofs << *b++ << std::endl;
	}
	ofs << std::flush;
	if (ofs.bad())
	{
		perror(OPTIONS);
		return (1);
	}

	if (mkdir(TMP, 0755))
	{
		perror(TMP);
		return (1);
	}

	if (mkdir(DOMAINS, 0755))
	{
		perror(DOMAINS);
		return (1);
	}

	if (mkdir(COMMANDS, 0755))
	{
		perror(COMMANDS);
		return (1);
	}
	if (mkdir(ARCHIVE, 0755))
	{
		perror(ARCHIVE);
		return (1);
	}
	if (mkdir(UNSUBLIST, 0755))
	{
		perror(UNSUBLIST);
		return (1);
	}
	if (mkdir(MODQUEUE, 0755))
	{
		perror(MODQUEUE);
		return (1);
	}
	if (mkdir(BOUNCES, 0755))
	{
		perror(BOUNCES);
		return (1);
	}

	return cmdupdate(argv);
}

// Set mailing list options

int cmdset(const std::vector<std::string> &argv)
{
	std::ofstream ofs(OPTIONS ".new");
	std::ifstream ifs(OPTIONS);

	if (!ofs.is_open() || !ifs.is_open())
	{
		perror(OPTIONS);
		return (1);
	}

	size_t i;

#if HAVE_SETLOCALE
	setlocale(LC_ALL, "");
#endif

	bool updateheaderadd=false;

	for (i=0; i<argv.size(); i++)
	{
		std::string n=argv[i];

		std::replace(n.begin(), n.end(), '\n', ' ');

		std::string::iterator b=n.begin(), e=n.end();
		std::string::iterator p=b;

		while (p != e)
		{
			if (*p & 0x80)
				break;
			++p;
		}

		std::string::iterator q=std::find(n.begin(), e, '=');

		if (q != e)
			++q;

		std::string w(q, e);

		std::string h_name(b, q);

		if (h_name == "ADDRESS=" ||
		    h_name == "URL=")
			updateheaderadd=true;

		if (p != e)
		{
			if (h_name != "NAME=")
			{
				const struct unicode_info *from=
					unicode_find(nl_langinfo(CODESET));
				char *conv_p;

				if (from &&
				    (conv_p=unicode_xconvert(w.c_str(), from,
							     &unicode_UTF8))
				    != NULL)
				{
					n=std::string(b,p)
						+ std::string(conv_p);
					free(conv_p);
				}
			}
			else
			{
				char *q=rfc2047_encode_str(w.c_str(),

#if HAVE_LANGINFO_H
							   nl_langinfo(CODESET),
#else
							   "UTF-8",
#endif
							   rfc2047_qp_allow_word);

				if (q)
				{
					n=std::string(b,p)+q;
					free(q);
				}
			}
		}

		ofs << n << std::endl;
	}
#if HAVE_SETLOCALE
	setlocale(LC_ALL, "C");
#endif

	std::string buf;

	while (!std::getline(ifs, buf).eof())
	{
		const char *p=buf.c_str();
		const char *q=strchr(p, '=');

		if (!q)	continue;

		for (i=0; i<argv.size(); i++)
		{
			const char *s=argv[i].c_str();

			if (strncmp(s, p, q-p) == 0 &&
				s[q-p] == '=')
				break;
		}
		if (i < argv.size())	continue;
		ofs << buf << std::endl;
	}

	if (ifs.bad() || ofs.bad())
	{
		perror(OPTIONS);
		return (1);
	}

	ofs.close();
	ifs.close();
	if (rename( OPTIONS ".new", OPTIONS))
	{
		perror(OPTIONS);
		return (1);
	}

	if (updateheaderadd)
		return updatelistheaders();

	return (0);
}

static int insertalias(std::string addr, std::string a);

int docmdalias(const char *addr, std::string subbuf)
{
	return insertalias(addr, header_s(subbuf, "x-alias"));
}

static int insertalias(std::string addr, std::string a)
{
	int	rc;

	rc=getinfo(addr, isfound);
	if (rc)
	{
		std::cerr << "<" << addr
			<< "> is not subscribed to this mailing list." << std::endl;
		return (9);
	}

	TrimLeft(a);
	TrimRight(a);

	std::string::iterator domain_p=std::find(a.begin(), a.end(), '@');


	std::transform(domain_p, a.end(), domain_p, std::ptr_fun(::tolower));

SubExclusiveLock subscription_lock;

DbObj	aliases;

	aliases.Open(ALIASES, "C");

	if (a.size() > 0)
	{
		std::string	key;

		key="1";
		key += a;

		if (aliases.Store(key, addr, "I"))
		{
			std::cerr << "<" << a << "> is an existing write-only alias."
				<< std::endl;
			return (9);
		}

		key="0";
		key += addr;

		std::string p=aliases.Fetch(key, "");

		if (p.size())
		{
			std::string s="1" + p;
			aliases.Delete(s);
		}

		aliases.Store(key, a, "R");
	}
	else
	{
		std::string key;

		key="0";
		key += addr;

		std::string p=aliases.Fetch(key, "");

		if (p.size())
		{
			std::string s="1"+p;

			aliases.Delete(s);
			aliases.Delete(key);
		}
	}
	return (0);
}

// List subscribed addresses

int cmdlsub(const std::vector<std::string> &args)
{
	std::string buf;
	SubscriberList subs;

	while (subs.Next(buf))
	{
		const char *p=subs.sub_info.c_str();
		time_t t=0;

		if (strncmp(p, "x-couriermlm-date:", 18) == 0)
		{
			p += 18;
			while (*p && isspace((int)(unsigned char)*p))
				++p;
			while (isdigit((int)(unsigned char)*p))
			{
				t=t * 10 + (*p-'0');
				++p;
			}
			if (t)
			{
				struct tm *tmptr=localtime(&t);
				char tbuf[100];

				if (strftime(tbuf, sizeof(tbuf),
					     "%d-%b-%Y", tmptr) > 0)
				{
					buf += " (";
					buf += tbuf;
					buf += ")";
				}
			}
		}

		std::cout << buf << std::endl;
	}
	return (0);
}

static void doexport(std::ostream &o)
{
	SubExclusiveLock subscription_lock;

	std::string buf;
	SubscriberList subs;

	while (subs.Next(buf))
	{
		o << buf << std::endl;

		std::istringstream i(subs.sub_info);

		std::string line;

		while (!std::getline(i, line).eof())
		{
			if (line.substr(0, 1) == ".")
				line="." + line;

			o << line << std::endl;
		}
		o << "." << std::endl;
	}

	SubscriberList aliaslist(true);

	while (aliaslist.Next(buf))
	{
		std::cout << "@" << buf << std::endl;
		std::cout << aliaslist.posting_alias << std::endl;
	}
}

int cmdexport(const std::vector<std::string> &args)
{
	doexport(std::cout);
	return (0);
}

static int doimport(std::istream &i)
{
	std::string	address;
	std::string	subinfo;
	std::string	linebuf;
	int	rc;

	while (!std::getline(i, address).eof())
	{
		subinfo="";

		if (address.substr(0, 1) == "@") // ALIAS
		{
			if (std::getline(i, subinfo).eof())
				break;

			address=address.substr(1);
			insertalias(address, subinfo);
			continue;
		}

		while (!std::getline(i, linebuf).eof())
		{
			if (linebuf == ".")	break;
			if (*linebuf.c_str() == '.')
				linebuf=linebuf.substr(1);
			subinfo += linebuf;
			subinfo += '\n';
		}

		rc=docmdsub(address.c_str(), subinfo, false);
		if (rc == 9)
		{
			std::cout << address << ": already subscribed." << std::endl;
			rc=0;
		}
		if (rc)	return (rc);
	}
	return (0);
}

int cmdimport(const std::vector<std::string> &args)
{
	return (doimport(std::cin));
}

int cmdlaliases(const std::vector<std::string> &args)
{
	std::string buf;
	SubscriberList subs(true);

	while (subs.Next(buf))
		std::cout << buf << ' ' << subs.posting_alias << std::endl;
	return (0);
}

// Display subscription information for this subscriber


static int print_info(std::string s)
{
	std::cout << s;
	return (0);
}

int cmdinfo(const std::vector<std::string> &args)
{
	int	rc=getinfo(args.empty() ? "":args[0], print_info);

	if (rc == EX_NOUSER)
		std::cerr << "No subscription information found." << std::endl;
	return (rc);
}
