/*
** Copyright 2007 Double Precision, Inc.  See COPYING for
** distribution information.
**
** $Id: webmlmd.C,v 1.4 2007/04/08 14:04:05 mrsam Exp $
*/

#include	"config.h"
#include	"cgi/cgi.h"
#include	"rfc822/rfc822.h"
#include	"numlib/numlib.h"
#include	"unicode/unicode.h"
#include	"datadir.h"
#include	"bindir.h"

#include	<stdio.h>
#include	<errno.h>
#include	<stdlib.h>
#if	HAVE_UNISTD_H
#include	<unistd.h>
#endif
#include	<string.h>
#include	<signal.h>

#include	<sys/types.h>
#include	<sys/stat.h>
#include	<sys/wait.h>
#if	HAVE_FCNTL_H
#include	<fcntl.h>
#endif
#if	HAVE_LOCALE_H
#if	HAVE_SETLOCALE
#include	<locale.h>
#endif
#endif

#include	<vector>
#include	<string>
#include	<algorithm>
#include	<iostream>
#include	<fstream>
#include	<map>

#define TEMPLATEDIR DATADIR "/couriermlm"

extern "C" {

	void fake_exit(int rc)
	{
		exit(rc);
	}

	void rfc2045_enomem(const char *errmsg)
	{
		cginocache();
		std::cout << "Content-Type: text/plain" << std::endl
			  << std::endl
			  << errmsg
			  << std::endl << std::flush;
		exit(1);
	}

	void error(const char *errmsg)
	{
		cginocache();
		printf("Content-Type: text/html; charset=us-ascii\n\n"
                       "<html><head><title>%s</title></head><body><h1>%s</h1></body></html>\n",
                       errmsg, errmsg);
		fake_exit(1);
	}
}

std::vector<std::string> mlm_dirs;

static const char slash[]="/";

inline std::string basename(std::string n)
{
	n=std::string(std::find_end(n.begin(), n.end(), slash, slash+1),
		      n.end());

	if (n.size() > 0)
		n=n.substr(1);
	return n;
}

template<typename T> T html_escape(T s)
{
	T t;

	t.reserve(s.size() * 6 / 5);
	class T::iterator b=s.begin(), e=s.end(), p=b;

	while (b != e)
	{
		if (*b == '&')
		{
			t += T(p, b);
			t.push_back('&');
			t.push_back('a');
			t.push_back('m');
			t.push_back('p');
			t.push_back(';');
			p= ++b;
			continue;
		}

		if (*b == '<')
		{
			t += T(p, b);
			t.push_back('&');
			t.push_back('l');
			t.push_back('t');
			t.push_back(';');
			p= ++b;
			continue;
		}

		if (*b == '>')
		{
			t += T(p, b);
			t.push_back('&');
			t.push_back('g');
			t.push_back('t');
			t.push_back(';');
			p= ++b;
			continue;
		}

		++b;
	}

	t += T(p, b);

	return t;
}

std::string toutf8str(std::wstring w)
{
	std::vector<unicode_char> u;

	u.reserve(w.size()+1);

	u.insert(u.end(), w.begin(), w.end());

	u.push_back(0);

	char *p=unicode_toutf8(&u[0]);

	std::string s;

	if (p)
	{
		s=p;
		free(p);
	}
	return s;
}

inline std::string html_escape(const char *a)
{
	return html_escape<std::string>(a);
}

static const char *self_url()
{
	const char *https=getenv("HTTPS");

	if (https && strcmp(https, "on") == 0)
		return cgihttpsscriptptr();

	return cgihttpscriptptr();
}

static std::string list_name()
{
	const char *pi=getenv("PATH_INFO");

	if (!pi)
		return "";

	if (*pi == '/')
		++pi;

	std::string list_name(pi);

	std::string::iterator b=list_name.begin(),
		p=std::find(b, list_name.end(), '/');

	return std::string(b, p);
}

static std::string extra_path()
{
	const char *pi=getenv("PATH_INFO");

	if (!pi)
		return "";

	if (*pi == '/')
		++pi;

	std::string list_name(pi);

	std::string::iterator b=list_name.begin(), e=list_name.end(),
		p=std::find(b, e, '/');

	if (p != e)
		++p;

	return std::string(p, e);
}

static int compar_last(const char *as, const char *bs)
{
	size_t al=strlen(as);
	size_t bl=strlen(bs);

	while (al && bl)
	{
		if (as[--al] != bs[--bl])
			return (int)(unsigned char)as[al] -
				(int)(unsigned char)bs[bl];

		if (as[al] == '/' && bs[bl] == '/')
			return 0;
	}

	return (int)al - (int)bl;
}

class OrderByListName {

public:
	OrderByListName();
	~OrderByListName();

	bool operator()(std::string, std::string) const;
};

OrderByListName::OrderByListName()
{
}

OrderByListName::~OrderByListName()
{
}

class EqualLastComponent {

public:
	EqualLastComponent();
	~EqualLastComponent();

	bool operator()(std::string, std::string) const;
};

EqualLastComponent::EqualLastComponent()
{
}

EqualLastComponent::~EqualLastComponent()
{
}

bool EqualLastComponent::operator()(std::string a, std::string b) const
{
	return compar_last(a.c_str(), b.c_str()) == 0;
}

class cmlm_filep {

	cmlm_filep(const cmlm_filep &); // UNDEFINED
	cmlm_filep &operator=(const cmlm_filep &); // UNDEFINED

public:
	FILE *fp;

	cmlm_filep();
	~cmlm_filep();
	void close();
};

cmlm_filep::cmlm_filep() : fp(NULL)
{
}

cmlm_filep::~cmlm_filep()
{
	close();
}

void cmlm_filep::close()
{
	if (fp)
		fclose(fp);
	fp=NULL;
}

bool run_couriermlm(std::string extension,
		    std::string sender,
		    std::string command,
		    std::vector<std::string> args,
		    pid_t &pid,
		    cmlm_filep &stdin_ret,
		    cmlm_filep &stdout_ret)
{
	int pipefd[2];

	if ((stdout_ret.fp=tmpfile()) == NULL)
		return false;

	if (pipe(pipefd) < 0)
	{
		return false;
	}

	pid=fork();

	if (pid < 0)
	{
		close(pipefd[0]);
		close(pipefd[1]);
		return false;
	}

	if (pid == 0)
	{
		if (dup2(pipefd[0], 0) < 0)
		{
			exit(1);
		}

		if (dup2(fileno(stdout_ret.fp), 1) < 0 || dup2(1, 2) < 0)
		{
			exit(1);
		}
		stdout_ret.close();
		fcntl(1, F_SETFD, 0);

		close(pipefd[0]);
		close(pipefd[1]);

		setenv("DEFAULT", extension.c_str(), 1);
		setenv("SENDER", sender.c_str(), 1);

		std::vector< std::vector<char> > argv_buf;

		{
			std::vector<char> v;

			v.insert(v.end(), command.begin(), command.end() );
			v.push_back(0);

			argv_buf.push_back(v);
		}

		{
			std::vector<char> v;

			v.push_back('.');
			v.push_back(0);

			argv_buf.push_back(v);
		}

		std::vector<std::string>::iterator ab=args.begin(),
			ae=args.end();

		while (ab != ae)
		{
			std::vector<char> v;

			v.push_back('.');
			v.push_back(0);

			v.insert(v.end(), ab->begin(), ab->end());
			v.push_back(0);

			argv_buf.push_back(v);
			++ab;
		}

		std::vector<char *> argv;

		argv.push_back(BINDIR "/couriermlm");

		std::vector< std::vector<char> >::iterator vb=argv_buf.begin(),
			ve=argv_buf.end();

		while (vb != ve)
		{
			argv.push_back( & (*vb)[0] );
			++vb;
		}

		argv.push_back(NULL);

		execv(argv[0], &argv[0]);
		perror(argv[0]);
		exit(1);
	}

	close(pipefd[0]);

	signal(SIGPIPE, SIG_IGN);
	stdin_ret.fp=fdopen(pipefd[1], "w");
	return true;
}

bool run_couriermlm(std::string extension,
		    std::string sender,
		    std::string command,
		    pid_t &pid,
		    cmlm_filep &stdin_ret,
		    cmlm_filep &stdout_ret)
{
	std::vector<std::string> args;

	return run_couriermlm(extension, sender, command, args, pid,
			      stdin_ret, stdout_ret);
}

bool check_couriermlm(cmlm_filep &filep, pid_t pid)
{
	int waitstat;
	pid_t p;

	filep.close();

	while ((p= ::wait(&waitstat)) != pid)
	{
		if (p == -1)
		{
			if (errno == EINTR)
				continue;
			return false;
		}
	}

	return waitstat == 0;
}

static void main2(void *);

static void init_dirs(void *dummy)
{
	const char *dirs=getenv("LISTS");
	const char *p, *q;

	p=q=dirs ? dirs:"";

	while(1)
	{
		if (*p == ':' || *p == 0)
		{
			std::string dirname=std::string(q, p);

			if (dirname.size() > 0)
				mlm_dirs.push_back(dirname);

			if (*p == 0)
				break;
			q=p+1;
		}

		++p;
	}

	if (mlm_dirs.size() == 0)
	{
		std::cerr << "Missing LISTS setting" << std::endl;
		exit(1);
	}

	std::sort(mlm_dirs.begin(), mlm_dirs.end(), OrderByListName());

	std::vector<std::string>::iterator
		b=mlm_dirs.begin(), e=mlm_dirs.end();

	b=std::adjacent_find(b, e, EqualLastComponent());

	if (b != e)
	{
		std::cerr <<
			"Two directories cannot have the same last component:"
			  << std::endl
			  << *b << std::endl
			  << b[1] << std::endl;
		exit(1);
	}

	b=mlm_dirs.begin();

	while (b != e)
	{
		struct stat stat_buf;

		if (stat(b->c_str(), &stat_buf) < 0)
		{
			perror(b->c_str());
			exit(1);
		}

		if (stat_buf.st_uid == 0)
		{
			std::cerr << *b << ": is owned by root!"
				  << std::endl;
			exit(1);
		}

		if (geteuid() == 0)
			libmail_changeuidgid(stat_buf.st_uid,
					     stat_buf.st_gid);

		if (geteuid() != stat_buf.st_uid ||
		    getegid() != stat_buf.st_gid)
		{
			std::cerr << "My uid/gid does not match "
				  << *b << "'s"
				  << std::endl;
			exit(1);
		}
		++b;
	}
}

int main(int argc, char **argv)
{
	const char *cmd=NULL;
	const char *port_s=getenv("PORT");
	const char *prefork_s=getenv("PREFORK");

	if (argc > 1)
		cmd=argv[1];

#if HAVE_SETLOCALE
	setlocale(LC_ALL, "C");
#endif

	/*
	** For convenience's sake, we need to start webmlmd by a helper script.
	** Rather than confusing folks with webmlmd.rc, automatically invoke
	** the helper script, when needed.
	*/

	if (cmd && strcmp(cmd, "check"))
	{
		std::vector<char> shscript_v;

		{
			std::string shscript=argv[0];

			shscript += ".rc";

			shscript_v.reserve(shscript.size()+1);

			shscript_v.insert(shscript_v.end(),
					  shscript.begin(),
					  shscript.end());
			shscript_v.push_back(0);
		}

		std::vector<char *> new_argv;

		new_argv.push_back(&shscript_v[0]);

		new_argv.insert(new_argv.end(), argv, argv+argc);
		new_argv.push_back(0);

		execvp(new_argv[0], &new_argv[0]);
		perror(new_argv[0]);
		exit(1);
	}

	if (!port_s || !*port_s)
	{
		std::cerr << "PORT not set in the configuration file"
			  << std::endl;
		exit(1);
	}

	if (argc > 1 && strcmp(argv[1], "check") == 0)
	{
		if (access(port_s, 0) == 0)
		{
			if (access(port_s, W_OK) < 0)
			{
				perror(port_s);
				exit(1);
			}
			exit(0);
		}

		if (mknod(port_s, S_IFSOCK | 0777, 0) < 0)
		{
			perror(port_s);
			exit(1);
		}
		exit(0);
	}

	if (cmd)
	{
		init_dirs(NULL);
		exit(0); /* The "check" command */
	}

	cgi_daemon(atoi(prefork_s), port_s, init_dirs, main2, NULL);

	return (0);
}

std::string getoption(std::string listdir,
		      std::string option)
{
	std::string n=listdir + "/options";

	std::ifstream i(n.c_str());

	std::string s;

	if (!i.is_open())
		return s;

	while (!std::getline(i, s).eof())
	{
		std::string::iterator b=s.begin(), e=s.end(),
			p=std::find(b, e, '=');

		if (std::string(b, p) == option)
		{
			if (p != e)
				++p;

			return std::string(p, e);
		}
	}
	return "";
}

std::string getoption(std::string option)
{
	return getoption(".", option);
}

std::wstring getwoption(std::string dir, std::string option)
{
	std::string s=getoption(dir, option);

	unicode_char *u=unicode_fromutf8(s.c_str());

	if (!u)
		return std::wstring();

	size_t ul=0;

	for (ul=0; u[ul]; ++ul)
		;

	std::wstring w(u, u+ul);
	free(u);

	return w;
}

std::wstring getwoption(std::string option)
{
	return getwoption(".", option);
}

std::string getlistname(std::string dir, std::string dirbasename)
{
	std::wstring listname=getwoption(dir, "LISTNAME");

	if (listname.size() == 0)
	{
		listname.insert(listname.end(),
				dirbasename.begin(),
				dirbasename.end());
	}

	return toutf8str(html_escape(listname));
}

std::string getlistname(std::string dirbasename)
{
	return getlistname(".", dirbasename);
}

bool OrderByListName::operator()(std::string a, std::string b) const
{
	return getlistname(a, basename(a)) < getlistname(b, basename(b));
}

static void emit_lists()
{
	std::vector<std::string>::iterator b=mlm_dirs.begin(),
		e=mlm_dirs.end();

	std::cout << "<ul>";

	while (b != e)
	{
		std::string n= basename(*b);

		std::cout << "<li><a href=\"" << self_url() << "/"
			  << html_escape(n) << "\">" << getlistname(*b, n)
			  << "</a></li>";
		++b;
	}
	std::cout << "</ul>";
}

static void emit_list_name()
{
	std::cout << getlistname(list_name());
}

static void emit_list_descr()
{
	std::cout << getoption("LISTDESCR");
}

static void emit_list_url()
{
	std::cout << self_url() << "/" << list_name();
}

static void showform(const char *filename,
		     const std::map<std::string, std::string> &parms)
{
	std::map<std::string, void (*)(void) > handler_list;

	handler_list["LISTS"]= &emit_lists;
	handler_list["LISTNAME"]= &emit_list_name;
	handler_list["LISTURL"]= &emit_list_url;
	handler_list["LISTDESCR"]= &emit_list_descr;

	std::ifstream i(filename);

	if (!i.is_open())
	{
		std::cout << filename << ": cannot open"
			  << std::endl;
		return;
	}

	std::string line;

	while (!std::getline(i, line).eof())
	{
		std::string::iterator b=line.begin(), e=line.end();

		while (b != e)
		{
			static const char mac_open[]="[@";
			static const char mac_close[]="@]";

			std::string::iterator p=
				std::search(b, e, mac_open, mac_open+2);

			std::cout << std::string(b, p);

			b=p;

			if (b == e)
				break;

			b += 2;

			p=std::search(b, e, mac_close, mac_close+2);

			std::string handler_name=std::string(b, p);

			b=p;
			if (b != e)
				b += 2;

			std::map<std::string, void (*)(void) >::iterator
				h=handler_list.find(handler_name);

			if (h != handler_list.end())
				(*h->second)();

			std::map<std::string, std::string>::const_iterator
				vb=parms.find(handler_name);

			if (vb != parms.end())
			{
				std::cout << html_escape(vb->second);
			}
		}
		std::cout << std::endl;
	}
}

static void showhtmlform(const char *formname,
			 std::map<std::string, std::string> &parms)
{
	cginocache();
	std::cout << "Content-Type: text/html; charset=utf-8"
		  << std::endl
		  << std::endl;

	showform(formname, parms);
}

static void showhtmlform(const char *formname)
{
	std::map<std::string, std::string> dummy;

	showhtmlform(formname, dummy);
}

static void showerrorform(const char *formname,
			  std::string errmsg)
{
	std::map<std::string, std::string> parms;

	parms["ERRMSG"]=errmsg;

	showhtmlform(formname, parms);
}


static void mk_received_header(FILE *fp)
{
	char buf[100];
	const char *ip=getenv("REMOTE_ADDR");

	rfc822_mkdate_buf(time(NULL), buf);

	fprintf(fp, "Received: from ([%s]) via http; %s\n",
		ip ? ip:"(unknown)", buf);
}

static void sendsubunsub(std::string ext)
{
	std::string address(cgi("address"));

	std::string::iterator b=address.begin(),
		e=address.end(),
		p=std::find(b, e, '@');

	if (p == e || std::find(p+1, e, '@') != e ||
	    std::find(p, e, '/') != e)
	{
		showerrorform("webmlmerror.tmpl.html",
			      "Invalid E-mail address: " + address);
		return;
	}

	pid_t pid;

	cmlm_filep cmlm_stdin;
	cmlm_filep cmlm_stdout;

	if (run_couriermlm(ext, address, "ctlmsg", pid, cmlm_stdin,
			   cmlm_stdout))
	{
		mk_received_header(cmlm_stdin.fp);
		fprintf(cmlm_stdin.fp, "\nSubscription request received.\n");

		if (check_couriermlm(cmlm_stdin, pid))
		{
			showhtmlform("webmlmrequestreceived.tmpl.html");
			return;
		}
	}

	showerrorform("webmlmerror.tmpl.html",
		      "An error occured processing this request.\n"
		      "Please try again later.\n");
}

static void listrequest2(std::string list_name,
			 std::string path_info)
{
	if (path_info == "sendsub")
	{
		sendsubunsub("subscribe");
		return;
	}

	if (path_info == "sendunsub")
	{
		sendsubunsub("unsubscribe");
		return;
	}

	if (path_info == "doconfirm")
	{
		std::string method(cgi("method"));
		std::string token(cgi("token"));

		std::string errmsg="Confirmation failed.  No further information is available.";

		if (std::find_if(method.begin(), method.end(),
				 std::not1(std::ptr_fun(::isalpha))) ==
		    method.end() &&
		    std::find_if(token.begin(), token.end(),
				 std::not1(std::ptr_fun(::isalpha))) ==
		    token.end())
		{
			cmlm_filep cmlm_stdin, cmlm_stdout;
			pid_t pid;

			if (run_couriermlm(method + "-" + token, "",
					   "ctlmsg", pid,
					   cmlm_stdin, cmlm_stdout))
			{
				mk_received_header(cmlm_stdin.fp);
				fprintf(cmlm_stdin.fp,
					"Subject: yes -- confirmed by WebMLM\n"
					"\n"
					"Confirmed\n");


				if (check_couriermlm(cmlm_stdin, pid))
				{
					showhtmlform("webmlmprocessed.html");
					return;
				}

				char buf[1024];
				bool isfirst=true;

				fseek(cmlm_stdout.fp, 0L, SEEK_END);
				fseek(cmlm_stdout.fp, 0L, SEEK_SET);

				while (fgets(buf, sizeof(buf), cmlm_stdout.fp))
				{
					if (isfirst)
					{
						isfirst=false;
						errmsg="";
					}
					errmsg += buf;
				}
			}
		}

		std::map<std::string, std::string> args;

		args["ERROR"]=html_escape(errmsg);
		showhtmlform("webmlmnotprocessed.html", args);
		return;
	}


	std::string::iterator b=path_info.begin(), e=path_info.end();

	std::string::iterator p=std::find(b, e, '/');

	std::string w(b, p);

	if (p != e) ++p;

	if (w == "subconfirm" || w == "unsubconfirm")
	{
		std::map<std::string, std::string> args;

		args["METHOD"]=w;
		args["TOKEN"]=std::string(p, e);
		showhtmlform("webmlmconfirm.html", args);
		return;
	}

	showhtmlform("webmlmlistindex.tmpl.html");
}

static void listrequest(std::string list_name,
			std::string path_info)
{
	std::vector<std::string>::iterator lb=mlm_dirs.begin(),
		le=mlm_dirs.end();

	while (lb != le)
	{
		std::string d=*lb++;

		if (basename(d) == list_name)
		{
			if (chdir(d.c_str()) < 0)
			{
				cginocache();
				std::cout << "Content-Type: text/plain"
					  << std::endl
					  << std::endl;
				perror(d.c_str());
				return;
			}
			listrequest2(list_name, path_info);
			return;
		}
	}

	cginocache();
	std::cout << "Content-Type: text/plain"
		  << std::endl
		  << std::endl
		  << "List " << list_name << " not found."
		  << std::endl;
	return;
}

static void main2(void *dummy)
{
	cgi_setup();

	std::string n=list_name();

	if (n.size() == 0)
	{
		showhtmlform(TEMPLATEDIR "/webmlmidx.html");
	}
	else
	{
		listrequest(n, extra_path());
	}

	std::cout.flush();
}
